from . import Message, Handler as Base

from bidict import bidict

import random
from abc import ABC


class PingPong(Message, ABC):
    NONCE_SIZE = 4 # bytes

    def __init__(self, nonce):
        self.nonce = nonce

    def __str__(self):
        return self.__repr__() + f'[nonce={self.nonce}]'

    def send(self, handler):
        handler.sock.sendall(
            self.nonce.to_bytes(self.NONCE_SIZE, handler.BYTE_ORDER))

    @classmethod
    def recv(cls, handler):
        nonce = handler.recv_bytes(cls.NONCE_SIZE)
        if nonce is None: return None
        nonce = int.from_bytes(nonce, handler.BYTE_ORDER)
        return cls(nonce)


class Ping(PingPong):
    def __init__(self, nonce=None):
        if nonce is None:
            nonce = random.randint(0, 256**self.NONCE_SIZE - 1)

        super().__init__(nonce)


class Pong(PingPong):
    def __init__(self, nonce):
        super().__init__(nonce)


class Handler(Base):
    MSG_BY_ID = bidict({
        1: Ping,
        2: Pong,
    })

    def __init__(self, sock, addr):
        super().__init__(sock, addr)

    def server(self):
        while msg := self.recv():
            print(f'     {self.addr} -> {msg}')

            if isinstance(msg, Ping):
                msg = Pong(msg.nonce)
                print(f'     {self.addr} <- {msg}')
                self.send(msg)






def main(args):
    print('LANK NODE: ', end='')

    cmd = args[0] if args else 'help'
    args = args[1:]

    if cmd in SET:
        cmd = SET[cmd]
        print(cmd[1])
        cmd[0](args)

    else:
        print(f'unknown command: {cmd}')
        print('Try `lank node help` for a list of commands.')


def help(args):
    print()
    print('USAGE:')
    print('   lank node <command>')
    print()
    print('The following commands are available:')
    for name, cmd in SET.items():
        print(f'   {name} - {cmd[1]}')


def run(args):
    from . import Server

    print('Starting server process...')
    server = Server()
    print(f'   listening on port {server.port}')
    server.run()


def test(args):
    from .test import TestClient
    tester = TestClient()
    print('Testing...')
    tester.run()


SET = {
    'help': (help,
        'help for node commands'),
    'run': (run,
        'run the node'),
    'test': (test,
        'test the node'),
}




from .config import DB

import sqlite3
from datetime import datetime


VERSION = 1

# meta table entries
META_VERSION = 'db_version'


#print(f'databasing "{DB}" ...')

con = sqlite3.connect(DB, isolation_level=None,
    detect_types=sqlite3.PARSE_DECLTYPES)
con.row_factory = sqlite3.Row
sqlite3.register_adapter(bool, int)
sqlite3.register_converter('BOOLEAN', lambda v: bool(int(v)))
cur = con.cursor()
#cur.execute('.dbconfig defensive on')
cur.execute('PRAGMA journal_mode=WAL')
cur.execute('PRAGMA synchronous=NORMAL')
cur.execute('PRAGMA temp_store=MEMORY')
#cur.execute('PRAGMA locking_mode=EXCLUSIVE')
cur.execute('PRAGMA foreign_keys=ON')


def cur_fetch():
    for row in cur:
        return row  # returns first row only

    return None  # if there were no rows

def cur_fetchcol(name):
    row = cur_fetch()
    if not row:
        return None

    return row[name]

def cur_fetchall():
    rows = [ ]

    for row in cur:
        rows.append(row)

    return rows

def close():
    cur.execute('VACUUM')
    con.close()


###
### META
###

cur.execute('''
    CREATE TABLE IF NOT EXISTS meta (
        name    TEXT PRIMARY KEY,
        value   TEXT
    )
''')

cur.executemany('''
    INSERT OR IGNORE INTO meta (
        name,
        value
    )
    VALUES (
        ?, ?
    )
''', [
    (META_VERSION, VERSION),
])

def get_meta(name):
    cur.execute('''
        SELECT value
        FROM meta
        WHERE name = ?
    ''', (
        name,
    ))

    return cur_fetchcol('value')

def set_meta(name, value):
    cur.execute('''
        UPDATE meta
        SET value = ?
        WHERE name = ?
    ''', (
        value,
        name
    ))

def _upgrade_if_needed_():
    v = int(get_meta(META_VERSION))

    if v < VERSION:
        from .patch import upgrade
        upgrade(DB, cur, cur_fetchall, set_meta, META_VERSION, v, VERSION)

_upgrade_if_needed_() # pause here and patch the db before continuing


###
### FOO
###



#print('db ready :-)')




from .cmd import main
import lank.db as ldb

import sys


def run():
    try:
        main(sys.argv[1:])

    except KeyboardInterrupt:
        pass

    finally:
        ldb.close()


if __name__ == '__main__':
    run()




from abc import ABC, abstractmethod


cache = { }


def get_handler(sock, addr):
    version = sock.recv(1)

    if version:
        version = version[0]

        if version not in cache:
            try:
                exec(f'from .v{version} import Handler as Protocol_v{version}')
                exec(f'cache[{version}] = Protocol_v{version}')

            except ModuleNotFoundError:
                cache[version] = None

        handler = cache[version]

        if not handler:
            raise ValueError(f'protocol version {version}')

        return handler(sock, addr)

    else:
        return None


class Handler(ABC):
    MSG_BY_ID = None # subclass defines (should be a bidict)
    ID_SIZE = 1 # bytes
    BYTE_ORDER = 'big'
    BUFFER_SIZE = 1024 # bytes

    def __init__(self, sock, addr):
        self.sock = sock
        self.addr = addr
        self.buffer = memoryview(bytearray(self.BUFFER_SIZE))

    def get_id_bytes(self, msg):
        try:
            mid = self.MSG_BY_ID.inverse[type(msg)]
            try:
                return mid.to_bytes(self.ID_SIZE, self.BYTE_ORDER)

            except OverflowError as e:
                raise ValueError(f'msg type id too big: {mid}') from e

        except KeyError as e:
            raise ValueError(f'unsupported message type: {type(msg)}') from e

    def get_msg_type(self, id_bytes):
        mid = int.from_bytes(id_bytes, self.BYTE_ORDER)
        try:
            return self.MSG_BY_ID[mid]

        except KeyError as e:
            raise ValueError(f'unsupported message type id: {mid}') from e

    def send(self, msg):
        self.sock.sendall(self.get_id_bytes(msg))
        msg.send(self)

    def recv(self):
        id_bytes = self.recv_bytes(self.ID_SIZE)
        if id_bytes is None: return None
        msg = self.get_msg_type(id_bytes)
        return msg.recv(self)

    def recv_bytes(self, size):
        assert size <= self.BUFFER_SIZE
        total = 0

        while total < size:
            read = self.sock.recv_into(self.buffer[total:], size - total)
            if not read: return None
            total += read

        return self.buffer[:total]

    @abstractmethod
    def server(self):
        raise NotImplemented()


class Message(ABC):
    def __init__(self):
        pass

    def send(self, handler):
        pass

    @classmethod
    def recv(cls, handler):
        return cls()






def main(args):
    print('LANK PEER: ', end='')

    cmd = args[0] if args else 'help'
    args = args[1:]

    if cmd in SET:
        cmd = SET[cmd]
        print(cmd[1])
        cmd[0](args)

    else:
        print(f'unknown command: {cmd}')
        print('Try `lank peer help` for a list of commands.')


def help(args):
    print()
    print('USAGE:')
    print('   lank peer <command>')
    print()
    print('The following commands are available:')
    for name, cmd in SET.items():
        print(f'   {name} - {cmd[1]}')


SET = {
    'help': (help,
        'help for peer commands'),
}




from appdirs import AppDirs

from os.path import join, exists
from os import makedirs


DIRS = AppDirs('lank', 'LANNOCC')

if not exists(DIRS.user_data_dir):
    makedirs(DIRS.user_data_dir)

DB = join(DIRS.user_data_dir, 'LANK.db')




from . import DEFAULT_PORT, HELLO
from .protocol.v1 import Handler, Ping, Pong

from gevent import socket

import sys


def begin(txt):
    print(f'   {txt}...', end='')
    sys.stdout.flush()

def end():
    print(' [done]')


class TestClient:
    def __init__(self):
        pass

    def run(self):
        addr = ('localhost', DEFAULT_PORT)

        begin('connecting')
        sock = socket.create_connection(addr)
        end()

        begin('sending HELLO v1')
        sock.sendall(HELLO + b'\x01')
        end()

        begin('instantiating protocol handler')
        handler = Handler(sock, addr)
        end()

        begin('ping-pong')
        for i in range(999):
            ping = Ping()

            print('>', end='')
            sys.stdout.flush()
            handler.send(ping)

            msg = handler.recv()
            print('<', end='')
            sys.stdout.flush()

            if isinstance(msg, Pong):
                if msg.nonce != ping.nonce:
                    print()
                    print(f'      NONCE: sent {ping.nonce}, got {msg.nonce}')
                    break

            else:
                print()
                print(f'      {msg}')
                break
        end()

